# Copyright 2016, 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import exceptions
from tempest.lib.services.compute import flavors_client
from tempest.lib.services.compute import hosts_client
from tempest.lib.services.compute import servers_client
from tempest.lib.services.identity.v2 import identity_client
from tempest.lib.services.identity.v2 import roles_client
from tempest.lib.services.identity.v2 import tenants_client
from tempest.lib.services.identity.v2 import users_client
from tempest.lib.services.identity.v3  \
    import identity_client as identity_v3_client
from tempest.lib.services.identity.v3 import projects_client
from tempest.lib.services.identity.v3 import roles_client as roles_v3_client
from tempest.lib.services.identity.v3 import services_client as s_client
from tempest.lib.services.identity.v3 import users_client as users_v3_client
from tempest.lib.services.image.v2 import images_client
from tempest.lib.services.network import networks_client
from tempest.lib.services.network import subnets_client
from tempest.lib.services.object_storage import account_client
try:
    # Since Rocky, volume.v3.services_client is the default
    from tempest.lib.services.volume.v3 import services_client
except ImportError:
    # For backward compatibility
    from tempest.lib.services.volume.v2 import services_client


class ProjectsClient(object):
    """The class is a wrapper for managing projects/tenants.

    It instantiates tempest projects_client and provides methods for creating
    and finding projects (identity version v3)/tenants (identity version v2).
    """
    def __init__(self, auth, catalog_type, identity_region, endpoint_type,
                 identity_version, **default_params):
        self.identity_version = identity_version
        if self.identity_version == "v2":
            self.project_class = tenants_client.TenantsClient
        else:
            self.project_class = projects_client.ProjectsClient
        self.client = self.project_class(auth, catalog_type, identity_region,
                                         endpoint_type, **default_params)

    def get_project_by_name(self, project_name):
        if self.identity_version == "v2":
            projects = self.client.list_tenants()['tenants']
        else:
            projects = self.client.list_projects()['projects']
        for project in projects:
            if project['name'] == project_name:
                return project
        raise exceptions.NotFound(
            'No such project/tenant (%s) in %s' % (project_name, projects))

    def create_project(self, name, description):
        if self.identity_version == "v2":
            self.client.create_tenant(name=name, description=description)
        else:
            self.client.create_project(name=name, description=description)


class ClientManager(object):
    """Manager of various OpenStack API clients.

    Connections to clients are created on-demand, i.e. the client tries to
    connect to the server only when it's being requested.
    """
    def __init__(self, conf, creds):
        """Init method of ClientManager.

        :param conf: TempestConf object
        :param creds: Credentials object
        """
        self.identity_region = creds.identity_region
        self.auth_provider = creds.get_auth_provider()

        default_params = creds.get_ssl_certificate_validation()
        compute_params = self._get_compute_params(conf)
        compute_params.update(default_params)

        catalog_type = conf.get_defaulted('identity', 'catalog_type')

        self.identity = self.get_identity_client(
            creds.identity_version,
            catalog_type,
            default_params)

        self.projects = ProjectsClient(
            self.auth_provider,
            conf.get_defaulted('identity', 'catalog_type'),
            self.identity_region,
            'publicURL',
            creds.identity_version,
            **default_params)

        self.set_roles_client(
            auth=self.auth_provider,
            identity_version=creds.identity_version,
            catalog_type=catalog_type,
            endpoint_type='publicURL',
            default_params=default_params)

        self.hosts_client = hosts_client.HostsClient(
            self.auth_provider,
            conf.get_defaulted('compute', 'catalog_type'),
            self.identity_region,
            **default_params)

        self.accounts = account_client.AccountClient(
            self.auth_provider,
            conf.get_defaulted('object-storage', 'catalog_type'),
            self.identity_region,
            **default_params)

        self.set_users_client(
            auth=self.auth_provider,
            identity_version=creds.identity_version,
            catalog_type=catalog_type,
            endpoint_type='publicURL',
            default_params=default_params)

        self.images = images_client.ImagesClient(
            self.auth_provider,
            conf.get_defaulted('image', 'catalog_type'),
            self.identity_region,
            **default_params)

        self.servers = servers_client.ServersClient(self.auth_provider,
                                                    **compute_params)
        self.flavors = flavors_client.FlavorsClient(self.auth_provider,
                                                    **compute_params)

        self.service_client = s_client.ServicesClient(
            self.auth_provider,
            conf.get_defaulted('identity', 'catalog_type'),
            self.identity_region,
            **default_params)

        self.volume_client = services_client.ServicesClient(
            self.auth_provider,
            conf.get_defaulted('volume', 'catalog_type'),
            self.identity_region,
            **default_params)

        self.networks = networks_client.NetworksClient(
            self.auth_provider,
            conf.get_defaulted('network', 'catalog_type'),
            self.identity_region,
            endpoint_type=conf.get_defaulted('network',
                                             'endpoint_type'),
            **default_params)

        self.subnets = subnets_client.SubnetsClient(
            self.auth_provider,
            conf.get_defaulted('network', 'catalog_type'),
            self.identity_region,
            endpoint_type=conf.get_defaulted('network',
                                             'endpoint_type'),
            **default_params)

        # Set admin project id needed for keystone v3 tests.
        if creds.admin:
            project = self.projects.get_project_by_name(creds.project_name)
            conf.set('auth', 'admin_project_id', project['id'])

    def _get_compute_params(self, conf):
        compute_params = {
            'service': conf.get_defaulted('compute', 'catalog_type'),
            'region': self.identity_region,
            'endpoint_type': conf.get_defaulted('compute', 'endpoint_type')
        }
        return compute_params

    def get_identity_client(self, identity_version, catalog_type,
                            default_params):
        """Obtain identity client.

        :type identity_version: string
        :type catalog_type: string
        :type default_params: dict
        """
        if "v3" in identity_version:
            return identity_v3_client.IdentityClient(
                self.auth_provider,
                catalog_type,
                self.identity_region, endpoint_type='publicURL',
                **default_params)
        else:
            return identity_client.IdentityClient(
                self.auth_provider,
                catalog_type,
                self.identity_region, endpoint_type='publicURL',
                **default_params)

    def get_service_client(self, service_name):
        """Returns name of the service's client.

        :type service_name: string
        :rtype: client object or None when the client doesn't exist
        """
        # TODO(arxcruz): This function is under used, it should return
        # a dictionary of all services for a particular client, for
        # example, we need hosts_client and flavors_client for compute
        # should return {'hosts': self.hosts_client, 'flavors': self.flavors }
        # and so on.
        if service_name == "image":
            return self.images
        elif service_name in ["network", "object-store"]:
            # return whole ClientManager object because NetworkService
            # currently needs to have an access to get_neutron/nova_client
            # methods which are chosen according to neutron presence
            return self
        elif service_name == "compute":
            return self.hosts_client
        elif "volume" in service_name:
            return self.volume_client
        elif service_name == "metering":
            return self.service_client
        elif service_name == "orchestration":
            return {'projects': self.projects,
                    'roles': self.roles,
                    'users': self.users,
                    'networks': self.networks,
                    'subnets': self.subnets}
        else:
            return None

    def set_users_client(self, auth, identity_version, catalog_type,
                         endpoint_type, default_params):
        """Sets users client.

        :param auth: auth provider
        :type auth: auth.KeystoneV2AuthProvider (or V3)
        :type identity_version: string
        :type catalog_type: string
        :type endpoint_type: string
        :type default_params: dict
        """
        users_class = users_client.UsersClient
        if "v3" in identity_version:
            users_class = users_v3_client.UsersClient

        self.users = users_class(
            auth,
            catalog_type,
            self.identity_region,
            endpoint_type=endpoint_type,
            **default_params)

    def set_roles_client(self, auth, identity_version, catalog_type,
                         endpoint_type, default_params):
        """Sets roles client.

        :param auth: auth provider
        :type auth: auth.KeystoneV2AuthProvider (or V3)
        :type identity_version: string
        :type catalog_type: string
        :type endpoint_type: string
        :type default_params: dict
        """
        roles_class = roles_client.RolesClient
        if "v3" in identity_version:
            roles_class = roles_v3_client.RolesClient

        self.roles = roles_class(
            auth,
            catalog_type,
            self.identity_region,
            endpoint_type=endpoint_type,
            **default_params)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from six.moves import configparser

from config_tempest.services import ceilometer
from config_tempest.tempest_conf import TempestConf
from config_tempest.tests.base import BaseServiceTest


class TestCeilometerService(BaseServiceTest):
    def setUp(self):
        super(TestCeilometerService, self).setUp()
        self.Service = ceilometer.MeteringService("ServiceName",
                                                  "ServiceType",
                                                  self.FAKE_URL,
                                                  self.FAKE_TOKEN,
                                                  disable_ssl_validation=False)
        self.conf = TempestConf()

    def test_check_ceilometer_service(self):
        client_service_mock = self.FakeServiceClient(services={})
        self.Service.client = client_service_mock
        self.Service.post_configuration(self.conf, client_service_mock)

        self._assert_conf_get_not_raises(configparser.NoSectionError,
                                         "service_available",
                                         "ceilometer")

        client_service_mock = self.FakeServiceClient(services={
            'services': [
                {
                    "name": "ceilometer",
                    "enabled": True
                }
            ]
        })
        self.Service.client = client_service_mock
        self.Service.post_configuration(self.conf, client_service_mock)
        self.Service.post_configuration(self.conf, client_service_mock)
        self.assertEqual(self.conf.get('service_available', 'ceilometer'),
                         'True')



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from config_tempest.tests.base import BaseConfigTempestTest
from config_tempest.users import Users
from tempest.lib import exceptions


class TestUsers(BaseConfigTempestTest):

    def setUp(self):
        # TODO(arxcruz): All these tests are running on identity v2 only, we
        # need to create tests for v3 too!
        # Story 2003388
        super(TestUsers, self).setUp()
        self.conf = self._get_conf("v2.0", "v3")
        self.conf.set("auth", "tempest_roles", "fake_role")
        projects_client = self._get_clients(self.conf).projects
        users_client = self._get_clients(self.conf).users
        roles_client = self._get_clients(self.conf).roles
        self.Service = Users(projects_client, roles_client,
                             users_client, self.conf)
        self.username = "test_user"
        self.password = "cryptic"
        self.project_name = "project"
        description = "Project for Tempest %s user" % self.username
        self.project_description = description
        self.role_name = "fake_role"
        self.email = "%s@test.com" % self.username
        self.users = {'users':
                      [{'name': "test_user",
                        'id': "fake_user_id"},
                       {'name': "test_user2",
                        'id': "fake_user_id2"}]}
        self.roles = {'roles':
                      [{'name': "fake_role",
                        'id': "fake_role_id"},
                       {'name': "fake_role2",
                        'id': "fake_role_id2"}]}

    @mock.patch('config_tempest.users.Users.'
                'create_user_with_project')
    @mock.patch('config_tempest.users.Users.give_role_to_user')
    def _test_create_tempest_user(self,
                                  mock_give_role_to_user,
                                  mock_create_user_with_project):
        alt_username = "my_user"
        alt_password = "my_pass"
        alt_project_name = "my_project"
        self.conf.set("identity", "alt_username", alt_username)
        self.conf.set("identity", "alt_password", alt_password)
        self.conf.set("identity", "alt_project_name", alt_project_name)
        self.Service.create_tempest_users()
        mock_give_role_to_user.assert_called_with(
            self.conf.get('auth', 'admin_username'),
            role_name='admin')
        self.assertEqual(mock_create_user_with_project.mock_calls, [
            mock.call(self.conf.get('identity', 'username'),
                      self.conf.get('identity', 'password'),
                      self.conf.get('identity', 'project_name')),
            mock.call(self.conf.get('identity', 'alt_username'),
                      self.conf.get('identity', 'alt_password'),
                      self.conf.get('identity', 'alt_project_name')),
        ])

    def test_create_tempest_user(self):
        self._test_create_tempest_user()

    @mock.patch('config_tempest.clients.ProjectsClient'
                '.get_project_by_name')
    @mock.patch('config_tempest.clients.ProjectsClient.create_project')
    @mock.patch('tempest.lib.services.identity.v2.users_client.'
                'UsersClient.create_user')
    def test_create_user_with_project(self,
                                      mock_create_user,
                                      mock_create_project,
                                      mock_get_project_by_name):
        mock_get_project_by_name.return_value = {'id': "fake-id"}
        self.Service.create_user_with_project(
            username=self.username,
            password=self.password,
            project_name=self.project_name)
        mock_create_project.assert_called_with(
            name=self.project_name, description=self.project_description)
        mock_create_user.assert_called_with(name=self.username,
                                            password=self.password,
                                            tenantId="fake-id",
                                            email=self.email)

    @mock.patch('config_tempest.clients.ProjectsClient'
                '.get_project_by_name')
    @mock.patch('config_tempest.clients.ProjectsClient.create_project')
    @mock.patch('tempest.lib.services.identity.v2'
                '.users_client.UsersClient.create_user')
    def test_create_user_with_project_project_exists(
            self,
            mock_create_user,
            mock_create_project,
            mock_get_project_by_name):
        mock_get_project_by_name.return_value = {'id': "fake-id"}
        exc = exceptions.Conflict
        mock_create_project.side_effect = exc
        self.Service.create_user_with_project(
            username=self.username,
            password=self.password,
            project_name=self.project_name)
        mock_create_project.assert_called_with(
            name=self.project_name, description=self.project_description)
        mock_create_user.assert_called_with(
            name=self.username,
            password=self.password,
            tenantId="fake-id",
            email=self.email)

    @mock.patch('tempest.common.identity.get_user_by_username')
    @mock.patch('config_tempest.clients.ProjectsClient.'
                'get_project_by_name')
    @mock.patch('tempest.lib.services.identity.v2.'
                'tenants_client.TenantsClient.create_tenant')
    @mock.patch('tempest.lib.services.identity.'
                'v2.users_client.UsersClient.create_user')
    def test_create_user_with_project_user_exists(
            self, mock_create_user, mock_create_project,
            mock_get_project_by_name,
            mock_get_user_by_username):
        mock_get_project_by_name.return_value = {'id': "fake-id"}
        exc = exceptions.Conflict
        mock_create_user.side_effect = exc
        fake_user = {'id': "fake_user_id"}
        mock_get_user_by_username.return_value = fake_user
        self.Service.create_user_with_project(
            username=self.username, password=self.password,
            project_name=self.project_name)
        mock_create_project.assert_called_with(
            name=self.project_name, description=self.project_description)
        mock_create_user.assert_called_with(name=self.username,
                                            password=self.password,
                                            tenantId="fake-id",
                                            email=self.email)

    @mock.patch('tempest.common.identity.get_user_by_username')
    @mock.patch('config_tempest.clients.ProjectsClient.'
                'get_project_by_name')
    @mock.patch('config_tempest.clients.ProjectsClient.create_project')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.create_user')
    def test_create_user_with_project_exists_user_exists(
            self, mock_create_user, mock_create_project,
            mock_get_project_by_name,
            mock_get_user_by_username):
        mock_get_project_by_name.return_value = {'id': "fake-id"}
        exc = exceptions.Conflict
        mock_create_project.side_effects = exc
        mock_create_user.side_effect = exc
        fake_user = {'id': "fake_user_id"}
        mock_get_user_by_username.return_value = fake_user
        self.Service.create_user_with_project(username=self.username,
                                              password=self.password,
                                              project_name=self.project_name)
        mock_create_project.assert_called_with(
            name=self.project_name, description=self.project_description)
        mock_create_user.assert_called_with(name=self.username,
                                            password=self.password,
                                            tenantId="fake-id",
                                            email=self.email)

    @mock.patch('config_tempest.clients.ProjectsClient.'
                'get_project_by_name')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.list_users')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.create_user')
    @mock.patch('tempest.lib.services.identity.v2.'
                'roles_client.RolesClient.list_roles')
    @mock.patch('tempest.lib.services.identity.v2.''roles_client.'
                'RolesClient.create_user_role_on_project')
    def test_give_role_to_user(self,
                               mock_create_user_role_on_project,
                               mock_list_roles,
                               mock_create_user,
                               mock_list_users,
                               mock_get_project_by_name):

        mock_get_project_by_name.return_value = \
            {'id': "fake_project_id"}
        mock_list_users.return_value = self.users
        mock_list_roles.return_value = self.roles
        self.Service.give_role_to_user(
            username=self.username,
            role_name=self.role_name)
        mock_create_user_role_on_project.assert_called_with(
            "fake_project_id", "fake_user_id", "fake_role_id")

    @mock.patch('config_tempest.clients.ProjectsClient.'
                'get_project_by_name')
    @mock.patch('tempest.lib.services.identity.'
                'v2.users_client.UsersClient.list_users')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.create_user')
    @mock.patch('tempest.lib.services.identity.v2.'
                'roles_client.RolesClient.list_roles')
    @mock.patch('tempest.lib.services.identity.v2.'
                'roles_client.RolesClient.create_user_role_on_project')
    def test_give_role_to_user_role_not_found(
            self,
            mock_create_user_role_on_project,
            mock_list_roles,
            mock_create_user,
            mock_list_users,
            mock_get_project_by_name):
        role_name = "fake_role_that_does_not_exist"
        mock_get_project_by_name.return_value = \
            {'id': "fake_project_id"}
        mock_list_users.return_value = self.users
        mock_list_roles.return_value = self.roles
        exc = Exception
        self.assertRaises(exc,
                          self.Service.give_role_to_user,
                          username=self.username,
                          role_name=role_name)

    @mock.patch('config_tempest.clients.ProjectsClient.'
                'get_project_by_name')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.list_users')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.create_user')
    @mock.patch('tempest.lib.services.identity.v2.'
                'roles_client.RolesClient.list_roles')
    @mock.patch('tempest.lib.services.identity.v2.roles_client'
                '.RolesClient.create_user_role_on_project')
    def test_give_role_to_user_role_not_found_not_req(
            self,
            mock_create_user_role_on_project,
            mock_list_roles,
            mock_create_user,
            mock_list_users,
            mock_get_project_by_name):

        mock_get_project_by_name.return_value = \
            {'id': "fake_project_id"}
        mock_list_users.return_value = self.users
        mock_list_roles.return_value = self.roles
        self.Service.give_role_to_user(
            username=self.username,
            role_name=self.role_name,
            role_required=False)

    @mock.patch('config_tempest.clients.ProjectsClient'
                '.get_project_by_name')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.list_users')
    @mock.patch('tempest.lib.services.identity.v2.'
                'users_client.UsersClient.create_user')
    @mock.patch('tempest.lib.services.identity.v2.'
                'roles_client.RolesClient.list_roles')
    @mock.patch('tempest.lib.services.identity.v2.roles_client.'
                'RolesClient.create_user_role_on_project')
    def test_give_role_to_user_role_already_given(
            self,
            mock_create_user_role_on_project,
            mock_list_roles,
            mock_create_user,
            mock_list_users,
            mock_get_project_by_name):
        exc = exceptions.Conflict
        mock_create_user_role_on_project.side_effect = exc
        mock_get_project_by_name.return_value = {'id': "fake_project_id"}
        mock_list_users.return_value = self.users
        mock_list_roles.return_value = self.roles
        self.Service.give_role_to_user(
            username=self.username,
            role_name=self.role_name)

    def _check_user_roles(self, user_roles, system_roles):
        self.Service._conf.set('auth', 'tempest_roles', user_roles)
        return self.Service.check_user_roles(system_roles)

    @mock.patch('logging.Logger.debug')
    def _check_user_role_does_not_exist(self, system_roles, LOG_mock,
                                        default_role='member'):
        roles = self._check_user_roles('doesNotExist', system_roles)
        # check if it fell down to member
        conf = self.Service._conf
        self.assertEqual(conf.get('auth', 'tempest_roles'), default_role)
        self.assertEqual(roles, [])
        self.assertEqual(len(LOG_mock.mock_calls), 3)

    def test_check_user_role_exists(self):
        system_roles = {'roles': [{'name': 'role1'}, {'name': 'role2'}]}
        roles = self._check_user_roles('role1', system_roles)
        self.assertEqual(roles[0], 'role1')

    @mock.patch('logging.Logger.debug')
    def test_check_user_roles_one_exists(self, LOG_mock):
        system_roles = {'roles': [{'name': 'role1'}, {'name': 'role2'}]}
        roles = self._check_user_roles('role1, doesNotExist', system_roles)
        self.assertEqual(roles[0], 'role1')
        self.assertEqual(len(LOG_mock.mock_calls), 2)

    @mock.patch('logging.Logger.debug')
    def test_check_user_roles_two_exist(self, LOG_mock):
        system_roles = {'roles': [{'name': 'role1'}, {'name': 'role2'}]}
        roles = self._check_user_roles('role1,role2', system_roles)
        self.assertEqual(roles[0], 'role1')
        self.assertEqual(roles[1], 'role2')
        self.assertEqual(len(LOG_mock.mock_calls), 1)

    def test_check_user_role_does_not_exist_fall_to_member(self):
        system_roles = {'roles': [{'name': 'role1'}, {'name': 'member'}]}
        self._check_user_role_does_not_exist(system_roles)

    def test_check_user_role_does_not_exist_fall_to_Member(self):
        system_roles = {'roles': [{'name': 'role1'}, {'name': 'Member'}]}
        self._check_user_role_does_not_exist(system_roles,
                                             default_role='Member')

    @mock.patch('logging.Logger.debug')
    def test_check_user_role_does_not_exist_no_member(self, LOG_mock):
        system_roles = {'roles': [{'name': 'role1'}]}
        roles = self._check_user_roles('doesNotExist', system_roles)
        self.assertEqual(roles, [])
        self.assertEqual(len(LOG_mock.mock_calls), 4)



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from tempest.lib import exceptions

from config_tempest import constants as C
from config_tempest.services.base import VersionedService


class ComputeService(VersionedService):
    def set_versions(self):
        url, top_level = self.no_port_cut_url()
        body = self.do_get(url, top_level=top_level)
        self.versions_body = json.loads(body)
        self.versions = self.deserialize_versions(self.versions_body)

    def set_default_tempest_options(self, conf):
        conf.set('compute-feature-enabled', 'console_output', 'True')
        num_compute = self._get_number_of_hosts()
        # set min_compute_nodes to enable migration tests in env
        # with multiple compute nodes
        conf.set('compute', 'min_compute_nodes', str(num_compute))
        # Resize only works if it has at least 2 compute nodes
        # or if nova has the option allow_resize_to_same_host
        # set to true. Unfortunately we can't get this info from
        # nova api, so we only set it when we know there's 2
        # compute nodes
        if num_compute >= 2:
            conf.set('compute-feature-enabled', 'resize', 'True')
        # set microversions
        m_versions = self.filter_api_microversions()
        conf.set('compute', 'min_microversion', m_versions['min_microversion'])
        conf.set('compute', 'max_microversion', m_versions['max_microversion'])

    def _get_number_of_hosts(self):
        # Right now the client returned is hosts, in the future
        # change it to a dict, and get the client as requested
        try:
            hosts = self.client.list_hosts()['hosts']
            compute_hosts = [h for h in hosts if h['service'] == 'compute']
            return len(compute_hosts)
        except exceptions.Forbidden:
            C.LOG.info('Can not retrieve hosts, user are not allowed')
            return 1

    def post_configuration(self, conf, is_service):
        conf.set('compute-feature-enabled', 'attach_encrypted_volume',
                 str(is_service(**{'type': 'key-manager'})))

    @staticmethod
    def get_service_type():
        return ['compute']

    @staticmethod
    def get_codename():
        return 'nova'



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from functools import wraps
import os
import shutil
import subprocess
import time

from six.moves import urllib
from tempest.lib import exceptions

from config_tempest import constants as C
from config_tempest.services.base import VersionedService


class ImageService(VersionedService):

    def set_image_preferences(self, disk_format, non_admin, no_rng=False,
                              convert=False):
        """Sets image prefferences.

        :type disk_format: string
        :type non_admin: bool
        :type no_rng: bool
        :type convert: bool
        """
        self.disk_format = disk_format
        self.non_admin = non_admin
        self.no_rng = no_rng
        self.convert = convert

    def set_default_tempest_options(self, conf):
        # When cirros is the image, set validation.image_ssh_user to cirros.
        # The option is heavily used in CI and it's also usefull for refstack,
        # because we don't have to specify overrides.
        if 'cirros' in conf.get_defaulted('image',
                                          'image_path').rsplit('/')[-1]:
            conf.set('validation', 'image_ssh_user', 'cirros')
        # image.http_image is a tempest option which defines 'http accessible
        # image', it can be in a compressed format so it can't be mistaken
        # for an image which will be uploaded to the glance.
        # image.http_image and image.image_path can be 2 different images.
        # If image.http_image wasn't set as an override, it will be set to
        # image.image_path or to DEFAULT_IMAGE
        image_path = conf.get_defaulted('image', 'image_path')
        if self._find_image_by_name(image_path) is None:
            conf.set('image', 'http_image', image_path)
        else:
            # image.image_path is name of the image already present in glance,
            # this value can't be set to image.http_image, therefor set the
            # default value
            conf.set('image', 'http_image', C.DEFAULT_IMAGE)

    def get_supported_versions(self):
        return ['v1', 'v2']

    @staticmethod
    def get_service_type():
        return ['image']

    @staticmethod
    def get_codename():
        return 'glance'

    def set_versions(self):
        super(ImageService, self).set_versions(top_level=False)

    def create_tempest_images(self, conf):
        """Uploads an image to the glance.

        The method creates images specified in conf, if they're not created
        already. Then it sets their IDs to the conf.

        :type conf: TempestConf object
        """
        img_dir = os.path.join(conf.get("scenario", "img_dir"))
        image_path = conf.get_defaulted('image', 'image_path')
        img_path = os.path.join(img_dir,
                                os.path.basename(image_path))
        name = image_path[image_path.rfind('/') + 1:]
        if self.convert and name[-4:] == ".img":
            name = name[:-4] + ".raw"
        if not os.path.exists(img_dir):
            try:
                os.makedirs(img_dir)
            except OSError:
                raise
        alt_name = name + "_alt"
        image_id = None
        if conf.has_option('compute', 'image_ref'):
            image_id = conf.get('compute', 'image_ref')
        image_id = self.find_or_upload_image(image_id, name,
                                             image_source=image_path,
                                             image_dest=img_path)
        alt_image_id = None
        if conf.has_option('compute', 'image_ref_alt'):
            alt_image_id = conf.get('compute', 'image_ref_alt')
        alt_image_id = self.find_or_upload_image(alt_image_id, alt_name,
                                                 image_source=image_path,
                                                 image_dest=img_path)
        # get name of the image_id
        image_id_name = self._find_image(image_id, '')['name']
        conf.set('scenario', 'img_file', image_id_name)
        conf.set('compute', 'image_ref', image_id)
        conf.set('compute', 'image_ref_alt', alt_image_id)

    def find_or_upload_image(self, image_id, image_name, image_source='',
                             image_dest=''):
        """If the image is not found, uploads it.

        :type image_id: string
        :type image_name: string
        :type image_source: string
        :type image_dest: string
        """
        image = self._find_image(image_id, image_name)

        if image:
            C.LOG.info("(no change) Found image '%s'", image['name'])
            path = os.path.abspath(image_dest)
            if not os.path.isfile(path):
                self._download_image(image['id'], path)
        else:
            C.LOG.info("Creating image '%s'", image_name)
            if image_source.startswith("http:") or \
               image_source.startswith("https:"):
                self._download_file(image_source, image_dest)
            else:
                try:
                    shutil.copyfile(image_source, image_dest)
                except IOError:
                    # let's try if this is the case when a user uses already
                    # existing image in glance which is not uploaded as *_alt
                    if image_name[-4:] == "_alt":
                        image = self._find_image(None, image_name[:-4])
                        if image:
                            path = os.path.abspath(image_dest)
                            if not os.path.isfile(path):
                                self._download_image(image['id'], path)
                    else:
                        raise IOError
            image = self._upload_image(image_name, image_dest)
        return image['id']

    def _find_image(self, image_id, image_name):
        """Find image by ID or name (the image client doesn't have this).

        :type image_id: string
        :type image_name: string
        """
        if image_id:
            try:
                return self.client.show_image(image_id)
            except exceptions.NotFound:
                pass
        return self._find_image_by_name(image_name)

    def _find_image_by_name(self, image_name):
        """Find image by name.

        :type image_name: string
        :return: Information in a dict about the found image
        :rtype: dict or None if image was not found
        """
        for x in self.client.list_images()['images']:
            if x['name'] == image_name:
                return x
        return None

    def _upload_image(self, name, path):
        """Upload image file from `path` into Glance with `name`.

        :type name: string
        :type path: string
        """
        if self.convert:
            path = self.convert_image_to_raw(path)

        C.LOG.info("Uploading image '%s' from '%s'",
                   name, os.path.abspath(path))
        if self.non_admin:
            visibility = 'community'
        else:
            visibility = 'public'

        with open(path, 'rb') as data:
            args = {'name': name, 'disk_format': self.disk_format,
                    'container_format': 'bare', 'visibility': visibility,
                    'hw_rng_model': 'virtio'}
            if self.no_rng:
                args.pop('hw_rng_model')
            image = self.client.create_image(**args)
            self.client.store_image_file(image['id'], data)
        return image

    def _download_image(self, id, path):
        """Download image from glance.

        :type id: string
        :type path: string
        """
        C.LOG.info("Downloading image %s to %s", id, path)
        body = self.client.show_image_file(id)
        C.LOG.debug(type(body.data))
        with open(path, 'wb') as out:
            out.write(body.data)

    def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
        """Retry calling the decorated function using exponential backoff

        http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
        original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

        Licensed under the BSD 3-Clause "New" or "Revised" License
        (https://github.com/saltycrane/retry-decorator/blob/master/LICENSE)

        :param ExceptionToCheck: the exception to check
        :type ExceptionToCheck: Exception or tuple
        :param tries: number of times before giving up
        :type type: int
        :param delay: initial delay between retries in seconds
        :type type: int
        :param backoff: backoff multiplier e.g. value of 2 will double the
            delay each retry
        :type backoff: int
        :param logger: logger to use. If None, print
        :type logger: logging. Logger instance
        """
        def deco_retry(f):
            @wraps(f)
            def f_retry(*args, **kwargs):
                mtries, mdelay = tries, delay
                while mtries > 1:
                    try:
                        return f(*args, **kwargs)
                    except ExceptionToCheck as e:
                        msg = "%s, Retrying in %d seconds." % (str(e), mdelay)
                        if logger:
                            logger.warning(msg)
                        else:
                            print(msg)
                        time.sleep(mdelay)
                        mtries -= 1
                        mdelay *= backoff
                return f(*args, **kwargs)
            return f_retry
        return deco_retry

    @retry(urllib.error.URLError, logger=C.LOG)
    def retry_urlopen(self, url):
        """Opens url using urlopen. If it fails, it will try again.

        :type url: string
        """
        return urllib.request.urlopen(url)

    def _download_file(self, url, destination):
        """Downloads a file specified by `url` to `destination`.

        :type url: string
        :type destination: string
        """
        if os.path.exists(destination):
            C.LOG.info("Image '%s' already fetched to '%s'.", url, destination)
            return
        C.LOG.info("Downloading '%s' and saving as '%s'", url, destination)
        f = self.retry_urlopen(url)
        data = f.read()
        with open(destination, "wb") as dest:
            dest.write(data)

    def convert_image_to_raw(self, path):
        """Converts given image to raw format.

        :type path: string
        :return: path of the converted image
        :rtype: string
        """
        head, tail = os.path.split(path)
        name = tail.rsplit('.', 1)[0] + '.raw'
        raw_path = os.path.join(head, name)
        # check if converted already
        if os.path.exists(raw_path):
            C.LOG.info("Image already converted in '%s'.", raw_path)
        else:
            C.LOG.info("Converting image '%s' to '%s'",
                       os.path.abspath(path), os.path.abspath(raw_path))
            rc = subprocess.call(['qemu-img', 'convert', path, raw_path])
            if rc != 0:
                raise Exception("Converting of the image has finished with "
                                "non-zero return code. The return code was "
                                "'%d'", rc)
        self.disk_format = 'raw'
        return raw_path



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from ssl import CertificateError

from six.moves import urllib

from config_tempest import constants as C


def configure_horizon(conf):
    """Derive the horizon URIs from the identity's URI."""
    uri = conf.get('identity', 'uri')
    u = urllib.parse.urlparse(uri)
    base = '%s://%s%s' % (u.scheme, u.netloc.replace(
        ':' + str(u.port), ''), '/dashboard')
    assert base.startswith('http:') or base.startswith('https:')
    has_horizon = True
    try:
        urllib.request.urlopen(base)
    except urllib.error.URLError:
        has_horizon = False
    except CertificateError as ex:
        C.LOG.info('Certificate Error while discovering Horizon: %s', (ex))
        has_horizon = False

    conf.set('service_available', 'horizon', str(has_horizon))
    if has_horizon:
        conf.set('dashboard', 'dashboard_url', base + '/')
        conf.set('dashboard', 'login_url', base + '/auth/login/')



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from config_tempest.services.base import VersionedService


class PlacementService(VersionedService):
    def set_versions(self):
        url, top_level = self.no_port_cut_url()
        body = self.do_get(url, top_level=top_level,
                           top_level_path='placement')
        self.versions_body = json.loads(body)

    def set_default_tempest_options(self, conf):
        # set microversions
        m_versions = self.filter_api_microversions(max_version='max_version')
        conf.set(
            'placement', 'min_microversion', m_versions['min_microversion'])
        conf.set(
            'placement', 'max_microversion', m_versions['max_microversion'])

    @staticmethod
    def get_service_type():
        return ['placement']

    @staticmethod
    def get_codename():
        return 'placement'



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from config_tempest.services.compute import ComputeService
from config_tempest.tempest_conf import TempestConf
from config_tempest.tests.base import BaseServiceTest


class TestComputeService(BaseServiceTest):
    def setUp(self):
        super(TestComputeService, self).setUp()
        self.Service = ComputeService("ServiceName",
                                      "ServiceType",
                                      self.FAKE_URL,
                                      self.FAKE_TOKEN,
                                      disable_ssl_validation=False)

    def test_set_get_extensions(self):
        exp_resp = []
        self._set_get_extensions(self.Service, exp_resp, self.FAKE_EXTENSIONS)

    def test_set_get_versions(self):
        exp_resp = ['v2.0', 'v2.1']
        self._set_get_versions(self.Service, exp_resp, self.FAKE_VERSIONS)

    @mock.patch('config_tempest.services.compute'
                '.ComputeService._get_number_of_hosts')
    def test_set_default_tempest_options(self, mock_get_number_of_hosts):
        mock_get_number_of_hosts.return_value = 2
        conf = TempestConf()
        self.Service.set_default_tempest_options(conf)
        self.assertEqual(
            conf.get('compute',
                     'min_compute_nodes'), '2')
        self.assertEqual(
            conf.get('compute-feature-enabled',
                     'resize'), 'True')
        self.assertEqual(
            conf.get('compute-feature-enabled',
                     'console_output'), 'True')
        mock_get_number_of_hosts.assert_called_once()



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This script is able to compare two ini files specified by arguments, f.e.:
#   $ ./compare-ini.py ./file1.ini ./file2.ini

# If some of the values are actually lists divided by commas, f.e.:
#
# [section]
# key = value1,value2,value3
#
# then this script compares these values regardless their order.

import six
from six.moves import configparser
import sys


def compare_ini_files(conf1, path1, conf2, path2):
    failed = 0
    for section in conf1.sections():
        if not conf2.has_section(section):
            sys.stderr.write("%s doesn't have %s section.\n" %
                             (path2, section))
            failed = 1
            continue
        for (key, value) in conf1.items(section):
            values1 = value.split(',')
            if len(values1) > 1:
                # the value contains a list of strings divided by commas,
                # e.g. api_extensions
                values2 = conf2.get(section, key).split(',')
                for i in values1:
                    if i not in values2:
                        sys.stderr.write("%s value not in %s.%s of %s\n" %
                                         (i, section, key, path2))
                        failed = 1
            else:
                val1 = conf1.get(section, key)
                if not conf2.has_option(section, key):
                    sys.stderr.write("%s doesn't have %s option in %s "
                                     "section.\n" % (path2, key, section))
                    failed = 1
                else:
                    val2 = conf2.get(section, key)
                    if not val1 == val2:
                        sys.stderr.write("%s in %s != %s in %s under %s.%s\n" %
                                         (val1, path1, val2,
                                          path2, section, key))
                        failed = 1
    return failed


if __name__ == '__main__':
    if six.PY3:
        conf1 = configparser.ConfigParser()
        conf2 = configparser.ConfigParser()
    else:
        conf1 = configparser.SafeConfigParser()
        conf2 = configparser.SafeConfigParser()

    conf1.read(sys.argv[1])
    conf2.read(sys.argv[2])

    # compare first file to the other one
    ret = compare_ini_files(conf1, sys.argv[1], conf2, sys.argv[2])

    # compare the other file to the first one
    ret2 = compare_ini_files(conf2, sys.argv[2], conf1, sys.argv[1])

    sys.exit(ret or ret2)



# -*- coding: utf-8 -*-

# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from fixtures import MonkeyPatch

from config_tempest.tests.base import BaseConfigTempestTest


class TestCredentials(BaseConfigTempestTest):

    def setUp(self):
        super(TestCredentials, self).setUp()
        self.conf = self._get_conf("v2.0", "v3")
        self.creds = self._get_creds(self.conf)
        self.creds_v2 = self._get_creds(self.conf, v2=True)

    def test_get_credential(self):
        # set conf containing the newer values (admin creds in auth section)
        self.creds._conf = self._get_alt_conf("v2.0", "v3")
        resp = self.creds.get_credential("username")
        self.assertEqual(resp, "demo")
        # set admin credentials
        self.creds.admin = True
        resp = self.creds.get_credential("username")
        self.assertEqual(resp, "admin")

    def test_get_identity_version_v2(self):
        resp = self.creds_v2._get_identity_version()
        self.assertEqual(resp, 'v2')

    def test_get_identity_version_v3(self):
        conf = self._get_conf("v3", "v3")  # uri has to be v3
        creds = self._get_creds(conf)
        resp = creds._get_identity_version()
        self.assertEqual(resp, 'v3')

    def test_get_creds_kwargs(self):
        expected_resp = {
            'username': 'demo',
            'password': 'secret',
            'project_name': 'demo'
        }
        self.assertEqual(self.creds_v2._get_creds_kwargs(), expected_resp)
        self.creds_v2.identity_version = 'v3'
        expected_resp = {
            'username': 'demo',
            'password': 'secret',
            'project_name': 'demo',
            'domain_name': 'Default',
            'user_domain_name': 'Default'
        }
        self.assertEqual(self.creds._get_creds_kwargs(), expected_resp)

    def test_set_credentials_v2(self):
        mock_function = mock.Mock()
        function2mock = 'config_tempest.credentials.auth.get_credentials'
        self.useFixture(MonkeyPatch(function2mock, mock_function))
        self.creds_v2.username = "name"
        self.creds_v2.password = "pass"
        self.creds_v2.project_name = "Tname"
        self.creds_v2.set_credentials()
        mock_function.assert_called_with(
            auth_url=None, fill_in=False, identity_version='v2',
            disable_ssl_certificate_validation='true',
            ca_certs=None, password='pass', project_name='Tname',
            username='name')

    def test_set_credentials_v3(self):
        mock_function = mock.Mock()
        function2mock = 'config_tempest.credentials.auth.get_credentials'
        self.useFixture(MonkeyPatch(function2mock, mock_function))
        self.creds.username = "name"
        self.creds.password = "pass"
        self.creds.project_name = "project_name"
        self.creds.identity_version = "v3"
        self.creds.set_credentials()
        mock_function.assert_called_with(
            auth_url=None, fill_in=False, identity_version='v3',
            disable_ssl_certificate_validation='true',
            ca_certs=None, password='pass',
            username='name',
            project_name='project_name',
            domain_name='Default',
            user_domain_name='Default')

    def test_get_auth_provider_keystone_v2(self):
        # check if method returns correct method - KeystoneV2AuthProvider
        mock_function = mock.Mock()
        # mock V2Provider, if other provider is called, it fails
        func2mock = 'config_tempest.credentials.auth.KeystoneV2AuthProvider'
        self.useFixture(MonkeyPatch(func2mock, mock_function))
        resp = self.creds_v2.get_auth_provider()
        self.assertEqual(resp, mock_function())
        # check parameters of returned function
        self.creds_v2.get_auth_provider()
        mock_function.assert_called_with(self.creds_v2.tempest_creds,
                                         'http://172.16.52.151:5000/v2.0',
                                         'true', None)

    def test_get_auth_provider_keystone_v3(self):
        # check if method returns KeystoneV3AuthProvider
        # make isinstance return True
        mockIsInstance = mock.Mock(return_value=True)
        self.useFixture(MonkeyPatch('config_tempest.credentials.isinstance',
                                    mockIsInstance))
        mock_function = mock.Mock()
        # mock V3Provider, if other provider is called, it fails
        func2mock = 'config_tempest.credentials.auth.KeystoneV3AuthProvider'
        self.useFixture(MonkeyPatch(func2mock, mock_function))
        resp = self.creds.get_auth_provider()
        self.assertEqual(resp, mock_function())
        # check parameters of returned function
        self.creds.get_auth_provider()
        mock_function.assert_called_with(self.creds.tempest_creds,
                                         'http://172.16.52.151:5000/v3',
                                         'true', None)



# Copyright 2016, 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests
from tempest.lib import auth

from config_tempest import utils


class Credentials(object):
    """Class contains all needed credentials.

    Wrapps credentials obtained from TempestConf object and Tempest
    credentialsfrom auth library.
    """
    def __init__(self, conf, admin, **kwargs):
        """Init method of Credentials.

        :type conf: TempestConf object
        :param admin: True if the user is admin, False otherwise
        :type admin: Boolean
        """
        self.admin = admin
        self._conf = conf
        self.verify = kwargs.get('verify', True)
        self.cert = kwargs.get('cert', None)
        self.username = self.get_credential('username')
        self.password = self.get_credential('password')
        self.project_name = self.get_credential('project_name')
        self.identity_version = self._get_identity_version()
        self.api_version = 3 if self.identity_version == "v3" else 2
        self.identity_region = self._conf.get_defaulted('identity', 'region')
        self.set_ssl_certificate_validation()
        self.set_credentials()

    def set_ssl_certificate_validation(self):
        # is there a specific CA bundle to use?
        # self.verify is either a boolean, in which case it controls whether
        # server's TLS certificates are verified, or a string, in which case
        # it is a path to a CA bundle to use, default in requests package
        # is True.
        if isinstance(self.verify, str):
            self.disable_ssl_certificate_validation = False
            self.ca_certs = self.verify
            self._conf.set('identity', 'ca_certificates_file', self.ca_certs)
        else:
            self.disable_ssl_certificate_validation = self._conf.get_defaulted(
                'identity', 'disable_ssl_certificate_validation'
            )
            self.ca_certs = self._conf.get_defaulted('identity',
                                                     'ca_certificates_file')
        self._conf.set('identity', 'disable_ssl_certificate_validation',
                       str(self.disable_ssl_certificate_validation))

    def get_ssl_certificate_validation(self):
        return {
            'disable_ssl_certificate_validation':
                self.disable_ssl_certificate_validation,
            'ca_certs': self.ca_certs,
        }

    def get_credential(self, key):
        """Helper for getting credential by its name.

        :param key: credential name
        :type key: string
        :returns: credential
        :rtype: string
        """
        if self.admin:
            # admin credentials are stored in auth section
            # and are prefixed by 'admin_'
            return self._conf.get_defaulted('auth', 'admin_' + key)
        else:
            # Tempest doesn't have non admin credentials, but the
            # tool keeps them in identity section for further usage
            return self._conf.get_defaulted('identity', key)

    def _list_versions(self, base_url, **kwargs):
        resp = requests.get(base_url, **kwargs)
        data = resp.json()
        return data["versions"]["values"]

    def _get_identity_version(self):
        """Looks for identity version in TempestConf object.

        :returns: identity version
        :rtype: string
        """
        base_url = utils.get_base_url(self._conf.get("identity", "uri"))
        kwargs = {
            'verify': self.verify,
            'cert': self.cert,
        }
        versions = self._list_versions(base_url, **kwargs)
        for version in versions:
            if version["status"] == "stable" and "v3" in version["id"]:
                return "v3"
        return "v2"

    def _get_creds_kwargs(self):
        """Creates kwargs.

        Kwargs based on the identity version, for obtaining
        Tempest credentials.

        :returns: kwargs
        :rtype: dict
        """
        creds_kwargs = {'username': self.username,
                        'password': self.password}
        if self.identity_version == 'v3':
            creds_kwargs.update({'project_name': self.project_name,
                                 'domain_name': 'Default',
                                 'user_domain_name': 'Default'})
        else:
            creds_kwargs.update({'project_name': self.project_name})
        return creds_kwargs

    def set_credentials(self):
        """Gets and saves Tempest credentials from auth library."""
        creds_kwargs = self._get_creds_kwargs()
        disable_ssl = self.disable_ssl_certificate_validation
        self.tempest_creds = auth.get_credentials(
            auth_url=None,
            fill_in=False,
            identity_version=self.identity_version,
            disable_ssl_certificate_validation=disable_ssl,
            ca_certs=self.ca_certs,
            **creds_kwargs)

    def get_auth_provider(self):
        """Gets auth provider based on the type of Tempest credentials.

        :returns: auth provider
        :rtype: auth.KeystoneV2AuthProvider/auth.KeystoneV3AuthProvider
        """
        if isinstance(self.tempest_creds, auth.KeystoneV3Credentials):
            # We set uri and uri_v3 to /v3 here because if the endpoint on the
            # rc file don't set the /v3 it will fail with a error 404
            uri = self._conf.get_defaulted('identity', 'uri_v3')
            uri = utils.get_base_url(uri) + 'v3'
            self._conf.set('identity', 'uri_v3', uri)
            return auth.KeystoneV3AuthProvider(
                self.tempest_creds,
                self._conf.get_defaulted('identity', 'uri_v3'),
                self.disable_ssl_certificate_validation,
                self.ca_certs)
        else:
            return auth.KeystoneV2AuthProvider(
                self.tempest_creds,
                self._conf.get_defaulted('identity', 'uri'),
                self.disable_ssl_certificate_validation,
                self.ca_certs)



# Copyright 2013, 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from config_tempest.constants import LOG
from config_tempest.services.base import VersionedService


class NetworkService(VersionedService):
    def set_extensions(self):
        body = self.do_get(self.service_url + '/v2.0/extensions.json')
        body = json.loads(body)
        self.extensions = list(map(lambda x: x['alias'], body['extensions']))

    def create_tempest_networks(self, conf, network_id):
        LOG.info("Setting up network")
        self.client = self.client.networks
        self.create_tempest_networks_neutron(conf, network_id)

    def create_tempest_networks_neutron(self, conf, public_network_id):
        self._public_network_name = None
        self._public_network_id = public_network_id
        # if user supplied the network we should use
        if public_network_id:
            self._supplied_network()
        # no network id provided, try to auto discover a public network
        else:
            self._discover_network()
        if self._public_network_id is not None:
            conf.set('network', 'public_network_id', self._public_network_id)
        if self._public_network_name is not None:
            conf.set('network', 'floating_network_name',
                     self._public_network_name)

    def get_service_extension_key(self):
        return 'api_extensions'

    def _supplied_network(self):
        LOG.info("Looking for existing network id: %s",
                 self._public_network_id)
        # check if network exists
        network_list = self.client.list_networks()
        for network in network_list['networks']:
            if network['id'] == self._public_network_id:
                self._public_network_name = network['name']
                break
        else:
            raise ValueError('provided network id: {0} was not found.'
                             ''.format(self._public_network_id))

    def _discover_network(self):
        LOG.info("No network supplied, trying auto discover for network")
        network_list = self.client.list_networks()
        for network in network_list['networks']:
            if network['router:external'] and network['subnets']:
                LOG.info("Found network, using: %s", network['id'])
                self._public_network_id = network['id']
                self._public_network_name = network['name']
                break

        # Couldn't find an existing external network
        else:
            LOG.error("No external networks found. "
                      "Please note that any test that relies on external "
                      "connectivity would most likely fail.")

    @staticmethod
    def get_service_type():
        return ['network']

    @staticmethod
    def get_codename():
        return 'neutron'



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from fixtures import MonkeyPatch

from config_tempest import constants as C
from config_tempest.flavors import Flavors
from config_tempest.tests.base import BaseConfigTempestTest


class TestFlavors(BaseConfigTempestTest):
    """Flavors test class

    Tests for create_tempest_flavors and find_or_create_flavor methods.
    """

    CLIENT_MOCK = 'tempest.lib.services.compute.flavors_client.FlavorsClient'
    FLAVORS_LIST = [
        {"id": "Fakeid", "name": "Name"},
        {"id": "MyFakeID", "name": "MyID"}
    ]

    def setUp(self):
        super(TestFlavors, self).setUp()
        self.conf = self._get_conf("v2.0", "v3")
        self.client = self._get_clients(self.conf).flavors
        return_value = {"flavors": [{"id": "MyFakeID", "name": "MyID"}]}
        mock_function = mock.Mock(return_value=return_value)
        self.useFixture(MonkeyPatch(self.CLIENT_MOCK + '.list_flavors',
                                    mock_function))
        self.Service = Flavors(self.client, True, self.conf, 128, 1)

    def test_create_tempest_flavors(self):
        self.Service.flavor_list = []
        mock_function = mock.Mock(return_value="FakeID")
        func2mock = 'config_tempest.flavors.Flavors.create_flavor'
        self.useFixture(MonkeyPatch(func2mock, mock_function))
        self.Service.create_tempest_flavors()
        self.assertEqual(self.conf.get('compute', 'flavor_ref'), "FakeID")
        self.assertEqual(self.conf.get('compute', 'flavor_ref_alt'), "FakeID")
        calls = [mock.call('m1.nano', 128, 1, 1, no_rng=False),
                 mock.call('m1.micro', 192, 1, 1, no_rng=False)]
        mock_function.assert_has_calls(calls, any_order=True)

    def check_call_of_discover_smallest_flavor(self):
        self.Service.flavor_list = [{'id': 'FAKE', 'name': 'Fake_flavor'},
                                    {'id': 'FAKE_1', 'name': 'Fake_flavor_1'}]
        self.Service.allow_creation = False
        func2mock = 'config_tempest.flavors.Flavors.discover_smallest_flavor'
        mock_function = mock.Mock()
        self.useFixture(MonkeyPatch(func2mock, mock_function))
        self.Service.create_flavor('nano', C.DEFAULT_FLAVOR_RAM,
                                   C.DEFAULT_FLAVOR_VCPUS,
                                   C.DEFAULT_FLAVOR_DISK)
        calls = [mock.call('nano')]
        mock_function.assert_has_calls(calls, any_order=True)

    def test_create_tempest_flavors_overwrite_flavor_ref_not_exist(self):
        self.conf.set('compute', 'flavor_ref', "FAKE_ID")
        try:
            self.Service.create_tempest_flavors()
        except Exception:
            return
        # it should have ended in the except block above
        self.assertTrue(False)

    def test_create_tempest_flavors_overwrite_flavor_ref_alt_not_exist(self):
        self.Service.flavor_list = [{'id': 'FAKE', 'name': 'Fake_flavor'}]
        self.conf.set('compute', 'flavor_ref', 'FAKE')
        self.conf.set('compute', 'flavor_ref_alt', 'FAKE_ID')
        try:
            self.Service.create_tempest_flavors()
        except Exception:
            self.assertEqual(self.conf.get('compute', 'flavor_ref'), 'FAKE')
            return
        # it should have ended in the except block above
        self.assertTrue(False)

    def test_create_flavor_not_allowed(self):
        # mock list_flavors() to return empty list
        self.Service.allow_creation = False
        self.Service.flavor_list = []
        try:
            self.Service.create_flavor('name', C.DEFAULT_FLAVOR_RAM,
                                       C.DEFAULT_FLAVOR_VCPUS,
                                       C.DEFAULT_FLAVOR_DISK)
        except Exception:
            return
        # it should have ended in the except block above
        self.assertTrue(False)

        # not enough flavors found
        self.Service.flavor_list = [{'id': 'FAKE', 'name': 'fake_name'}]
        try:
            self.Service.create_flavor('name', C.DEFAULT_FLAVOR_RAM,
                                       C.DEFAULT_FLAVOR_VCPUS,
                                       C.DEFAULT_FLAVOR_DISK)
        except Exception:
            return
        # it should have ended in the except block above
        self.assertTrue(False)

    def _test_create_flavor(self, no_rng=False):
        return_value = {"flavor": {"id": "MyFakeID", "name": "MyID"}}
        self.Service.flavor_list = []
        client = self.CLIENT_MOCK
        mock_function = mock.Mock(return_value=return_value)
        self.useFixture(MonkeyPatch(client + '.create_flavor', mock_function))
        mock_function = mock.Mock(return_value={})
        self.useFixture(MonkeyPatch(client + '.set_flavor_extra_spec',
                                    mock_function))
        resp = self.Service.create_flavor("MyID", C.DEFAULT_FLAVOR_RAM,
                                          C.DEFAULT_FLAVOR_VCPUS,
                                          C.DEFAULT_FLAVOR_DISK,
                                          no_rng=no_rng)
        self.assertEqual(resp, return_value['flavor']['id'])
        return mock_function

    def test_create_flavor_rng(self):
        mock_function = self._test_create_flavor()
        calls = [mock.call(flavor_id='MyFakeID', **{'hw_rng:allowed': 'True'})]
        mock_function.assert_has_calls(calls, any_order=True)

    def test_create_flavor_no_rng(self):
        self.Service.no_rng = True
        mock_function = self._test_create_flavor(no_rng=True)
        calls = [mock.call(flavor_id='MyFakeID')]
        mock_function.assert_has_calls(calls, any_order=True)

    def test_find_flavor_by_id(self):
        return_value = {"flavors": self.FLAVORS_LIST}
        mock_function = mock.Mock(return_value=return_value)
        self.useFixture(MonkeyPatch(self.CLIENT_MOCK + '.list_flavors',
                                    mock_function))
        resp = self.Service.find_flavor_by_id("MyFakeID")
        self.assertEqual(resp, "MyFakeID")
        # test no flavor found case
        resp = self.Service.find_flavor_by_id("NotExist")
        self.assertEqual(resp, None)

    def test_find_flavor_by_name(self):
        return_value = {"flavors": self.FLAVORS_LIST}
        mock_function = mock.Mock(return_value=return_value)
        self.useFixture(MonkeyPatch(self.CLIENT_MOCK + '.list_flavors',
                                    mock_function))
        resp = self.Service.find_flavor_by_name("MyID")
        self.assertEqual(resp, "MyFakeID")
        # test no flavor found case
        resp = self.Service.find_flavor_by_name("NotExist")
        self.assertEqual(resp, None)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from config_tempest import main
from config_tempest.services.object_storage import ObjectStorageService
from config_tempest import tempest_conf
from config_tempest.tests.base import BaseServiceTest


class TestObjectStorageService(BaseServiceTest):
    def setUp(self):
        super(TestObjectStorageService, self).setUp()
        self.Service = ObjectStorageService("ServiceName",
                                            "ServiceType",
                                            self.FAKE_URL,
                                            self.FAKE_TOKEN,
                                            disable_ssl_validation=False)
        self.Service.conf = tempest_conf.TempestConf()

    def test_set_get_extensions(self):
        expected_resp = ['formpost', 'ratelimit',
                         'methods', 'account_quotas']
        self._fake_service_do_get_method(self.FAKE_STORAGE_EXTENSIONS)
        self.Service.set_extensions()
        self.assertItemsEqual(self.Service.extensions, expected_resp)
        self.assertItemsEqual(self.Service.get_extensions(), expected_resp)

    def test_set_get_extensions_empty(self):
        self.Service.service_url = self.FAKE_URL + 'v3'
        self.Service.set_extensions()
        self.assertItemsEqual(self.Service.extensions, [])
        self.assertItemsEqual(self.Service.get_extensions(), [])

    def test_list_create_roles(self):
        conf = self.Service.conf
        # load default values
        main.load_basic_defaults(self.Service.conf)
        client = mock.Mock()
        return_mock = mock.Mock(return_value=self.FAKE_ROLES)
        client.list_roles = return_mock
        client.create_role = mock.Mock()
        self.Service.list_create_roles(conf, client)
        self.assertEqual(conf.get('object-storage', 'reseller_admin_role'),
                         'ResellerAdmin')
        self.assertEqual(conf.get('object-storage', 'operator_role'),
                         'admin')
        self.assertTrue(client.create_role.called)

    def test_check_service_status_discover(self):
        self.Service.client = mock.Mock()
        self.Service.client.accounts = mock.Mock()
        return_mock = mock.Mock(return_value=self.FAKE_ACCOUNTS)
        self.Service.client.accounts.skip_check = mock.Mock()
        self.Service.client.accounts.get = return_mock
        resp = self.Service.check_service_status(self.Service.conf)
        self.assertTrue(resp)

    def test_check_service_status(self):
        # discoverability set to False (e.g. via overrides)
        self.Service.conf.set('object-storage-feature-enabled',
                              'discoverability',
                              str(False))
        resp = self.Service.check_service_status(self.Service.conf)
        self.assertFalse(resp)
        # discoverability set to True (e.g. via overrides)
        self.Service.conf.set('object-storage-feature-enabled',
                              'discoverability',
                              str(True))
        resp = self.Service.check_service_status(self.Service.conf)
        self.assertTrue(resp)



# Copyright 2013, 2016, 2018 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import importlib
import pkgutil
import pyclbr

from six.moves import urllib

from config_tempest import constants as C
from config_tempest.services import horizon
from tempest.lib import exceptions

import config_tempest.services


class Services(object):
    def __init__(self, clients, conf, creds):
        self._clients = clients
        self._conf = conf
        self._creds = creds
        self._ssl_validation = creds.disable_ssl_certificate_validation
        self._region = clients.identity_region
        self._services = []
        self._service_classes = []
        self.set_catalog_and_url()
        self.available_services = self.get_available_services()

        self.discover()

    @property
    def service_classes(self):
        """Return the list of classes available under config_tempest.services.

        This return the list of classes that inherit from base.Service
        """
        if not self._service_classes:
            path = config_tempest.services.__path__
            prefix = config_tempest.services.__name__ + '.'
            for importer, modname, ispkg in pkgutil.walk_packages(
                    path=path, prefix=prefix, onerror=lambda x: None):
                module_info = pyclbr.readmodule(modname)
                for item in module_info.values():
                    m = importlib.import_module(modname)
                    c = getattr(m, item.name)
                    if issubclass(c, config_tempest.services.base.Service):
                        self._service_classes.append(c)

        return self._service_classes

    def get_available_services(self):
        try:
            services = self._clients.service_client.list_services()['services']
        except exceptions.Forbidden:
            C.LOG.warning("User has no permissions to list services, using "
                          "catalog. Services without endpoint will not be "
                          "discovered.")
            services = self.catalog
        return services

    def get_service_data(self, s_name, s_type):
        for s in self.catalog:
            if s['name'] == s_name and s['type'] == s_type:
                return s
        return None

    def discover(self):
        # We loop through the classes we have for each service, and if we find
        # a class that match a service enabled, we add it in our services list.
        # some services doesn't have endpoints, so we need to check first
        for s_class in self.service_classes:
            s_types = s_class.get_service_type()
            for s_type in s_types:
                s_name = [t['name'] for t in self.available_services
                          if t['type'] == s_type]
                if s_name:
                    # In the general case, there should only be one service in
                    # a deployment per service type
                    # https://docs.openstack.org/keystone/latest/contributor/
                    # service-catalog.html#services
                    if len(s_name) > 1:
                        C.LOG.warning("There are more service names ('%s') for"
                                      " '%s' service type, which is undefined"
                                      " behavior. Continuing with '%s'.",
                                      str(s_name), s_type, s_name[0])
                    s_name = s_name[0]
                    service_data = self.get_service_data(s_name, s_type)
                    url = None
                    if not service_data:
                        C.LOG.warning('No endpoint data found for %s', s_name)
                    else:
                        url = self.parse_endpoints(self.get_endpoints(
                            service_data), s_type)

                    # Create the service class and add it to services list
                    service = s_class(s_name, s_type, url, self.token,
                                      self._ssl_validation,
                                      self._clients.get_service_client(
                                          s_type),
                                      ca_certs=self._creds.ca_certs)
                    # discover extensions of the service
                    service.set_extensions()

                    # discover versions of the service
                    service.set_versions()
                    self.merge_exts_multiversion_service(service)

                    # default tempest options
                    service.set_default_tempest_options(self._conf)

                    service.set_availability(self._conf, True)

                    self._services.append(service)
                else:
                    # service is not available
                    # quickly instantiate a class in order to set
                    # availability of the service
                    s = s_class(None, None, None, None, None)
                    s.set_availability(self._conf, False)

    def merge_exts_multiversion_service(self, service):
        """Merges extensions of a service given by its name

        Looking for extensions from all versions of the service
        defined by name and merges them to that provided service.

        :param service: Service object
        """
        versions = service.get_supported_versions()
        service_type = service.get_unversioned_service_type()
        services_lst = []
        for v in versions:
            if self.is_service(**{'type': service_type + v}):
                services_lst.append(self.get_service(service_type + v))
        services_lst.append(service)
        service.extensions = self.merge_extensions(services_lst)

    def get_endpoints(self, entry):
        for ep in entry['endpoints']:
            if self._creds.api_version == 3:
                if (ep['region'] == self._region and
                        ep['interface'] == 'public'):
                    return ep
            else:
                if ep['region'] == self._region:
                    return ep
        try:
            return entry['endpoints'][0]
        except IndexError:
            return []

    def set_catalog_and_url(self):
        if self._creds.api_version == 3:
            service_catalog = 'catalog'
            self.public_url = 'url'
        else:
            service_catalog = 'serviceCatalog'
            self.public_url = 'publicURL'
        self.token, auth_data = self._clients.auth_provider.get_auth()
        self.catalog = auth_data[service_catalog]

    def parse_endpoints(self, ep, name):
        """Parse an endpoint(s).

        :param ep: endpoint(s)
        :type ep: dict or list in case of no endpoints
        :param name: name of a service
        :type name: string
        :return: url
        :rtype: string
        """
        # endpoint list can be empty
        if len(ep) == 0:
            url = ""
            C.LOG.info("Service %s has no endpoints", name)
        else:
            url = ep[self.public_url]
        if 'identity' in urllib.parse.urlparse(url).path:
            url = self.edit_identity_url(ep[self.public_url])
        return url

    def edit_identity_url(self, url):
        """A port and identity version are added to url if contains 'identity'

        :param url: url address of an endpoint
        :type url: string
        :rtype: string
        """
        # self._clients.auth_provider.auth_url stores identity.uri(_v3) value
        # from TempestConf
        port = urllib.parse.urlparse(self._clients.auth_provider.auth_url).port
        if port is None:
            port = ""
        else:
            port = ":" + str(port)
        replace_text = port + "/identity/" + self._creds.identity_version
        return url.replace("/identity", replace_text)

    def get_service(self, s_type):
        """Finds and returns a service object

        :param s_type: Type of a service
        :type s_type: string
        :return: Service object
        """
        for service in self._services:
            if service.s_type == s_type:
                return service
        return None

    def is_service(self, **kwargs):
        """Returns true if a service is available, false otherwise

        :param kwargs: Search parameters (accepts service name or type)
        :rtype: boolean
        """
        a_s = self.available_services
        if kwargs.get('name'):
            if [s for s in a_s if s['name'] == kwargs.get('name')]:
                return True
            return False

        if kwargs.get('type'):
            if [s for s in a_s if s['type'] == kwargs.get('type')]:
                return True
            return False
        return False

    def post_configuration(self):
        for s in self._services:
            s.post_configuration(self._conf, self.is_service)

        horizon.configure_horizon(self._conf)

    def set_supported_api_versions(self):
        # set supported API versions for services with more of them
        for service in self._services:
            versions = service.get_versions()
            supported_versions = service.get_supported_versions()
            if versions:
                section = service.get_feature_name() + '-feature-enabled'
                for s_version in supported_versions:
                    is_supported = any(s_version in item
                                       for item in versions)
                    self._conf.set(
                        section, 'api_' + s_version, str(is_supported))

    def merge_extensions(self, service_objects):
        """Merges extensions from all provided service objects

        :param service_objects:
        :type service_objects: list
        :return: Merged extensions
        :rtype: list
        """
        extensions = []
        for o in service_objects:
            if o:
                extensions += o.extensions
        return extensions

    def set_service_extensions(self):
        postfix = "-feature-enabled"
        try:
            keystone_v3_support = self._conf.getboolean('identity' + postfix,
                                                        'api_v3')
        except ValueError:
            keystone_v3_support = False
        if keystone_v3_support:
            self.get_service('identity').set_identity_v3_extensions()

        for service in self._services:
            ext_key = service.get_service_extension_key()
            if ext_key:
                extensions = ','.join(service.get_extensions())
                service_name = service.get_feature_name()
                self._conf.set(service_name + postfix, ext_key, extensions)
